@@ -3,7 +3,7 @@ import logging
 import re
 import shlex
 from asyncio import get_event_loop, TimeoutError
-from discord import FFmpegPCMAudio, PCMVolumeTransformer, Embed, opus, ClientException
+from discord import FFmpegPCMAudio, PCMVolumeTransformer, Embed, ClientException
 from math import floor, ceil
 from functools import partial
 from os import remove as remove_file
@@ -103,6 +103,7 @@ class MusicPlayer(BasePlugin):
                 and not self.config_manager.is_maintainer(msg.author):
             raise UserPermissionError
         player.queue.clear()
+        player.already_played.clear()
         player.stop()
         await player.voice_client.disconnect()
         del self.players[msg.guild.id]
@@ -126,8 +127,7 @@ class MusicPlayer(BasePlugin):
             urls = shlex.split(msg.clean_content)[1:]
         except ValueError as e:
             raise CommandSyntaxError(e)
-        for url in urls:
-            await player.enqueue(url)
+        await player.enqueue(urls)
 
     @Command("SongQueue", "Queue",
              doc="Tells the bot to list the current song queue.",
@@ -159,21 +159,20 @@ class MusicPlayer(BasePlugin):
         desc = vid.get("description", "*No description.*")
         if len(desc) > 2048:
             desc = desc[:2045] + "..."
-        embed = Embed(title=vid["title"], description=vid.get("description", "*No description.*"),
-                      url=vid["url"])
-        if "thumbnail" in vid:
+        embed = Embed(title=vid.get("title", "Unknown"), description=desc, url=vid["url"])
+        if vid.get("thumbnail") is not None:
             embed.set_thumbnail(url=vid["thumbnail"])
-        if "uploader" in vid:
+        if vid.get("uploader") is not None:
             embed.set_author(name=vid["uploader"])
-        if "tags" in vid:
+        if vid.get("tags") is not None:
             embed.set_footer(text=f"Tags: {', '.join(vid['tags'])}")
         play_time, duration, _ = player.progress
         rating_field = f"Views: {vid.get('view_count', 'Unknown'):,}."
-        if "like_count" in vid:
+        if vid.get("like_count") is not None:
             rating_field += f" {vid['like_count']:,}👍"
-        if "dislike_count" in vid:
+        if vid.get("dislike_count") is not None:
             rating_field += f"/{vid['dislike_count']:,}👎"
-        if vid.get("average_rating", None) is not None:
+        if vid.get("average_rating") is not None:
             rating_field += f" (Average rating: {vid['average_rating']:.2f})"
         embed.add_field(name="Ratings", value=rating_field)
         await respond(msg, embed=embed)
@@ -204,6 +203,14 @@ class MusicPlayer(BasePlugin):
         index -= 1
         try:
             del_song = player.queue.pop(index)
+            if index in player.already_played:
+                player.already_played.remove(index)
+            new_already_played = set()
+            for i in player.already_played:
+                if i > index:
+                    i -= 1
+                new_already_played.add(i)
+            player.already_played = new_already_played
         except IndexError:
             raise CommandSyntaxError("Integer provided is not a valid index")
         await respond(msg, f"**AFFIRMATIVE. Deleted song at position {index + 1} ({del_song['title']}).**")
@@ -212,27 +219,40 @@ class MusicPlayer(BasePlugin):
                 await respond(msg, split_msg)
 
     @Command("SongVolume", "Volume",
-             doc="Sets the volume at which the player plays music, between 0 and 100.",
+             doc="Sets the volume at which the player plays music, between 0 and 100.\nUse --temporary flag to "
+                 "automatically reset volume at the end of the current song.",
+             syntax="[-t/--temporary] (0-100)",
              category="music_player")
     async def _set_volume(self, msg):
         player = self.get_guild_player(msg)
         self.check_user_permission(msg.author, player)
+        new_volume = 0
         try:
-            new_volume = msg.clean_content.split(None, 1)[1]
-            new_volume = int(new_volume)
+            temp_flag = False
+            new_volume = msg.clean_content.split(None, 2)[1:]
+            if len(new_volume) > 1:
+                if new_volume[0] in ("-t", "--temporary"):
+                    temp_flag = True
+                    new_volume = int(new_volume[1])
+                else:
+                    raise CommandSyntaxError
+            else:
+                new_volume = int(new_volume[0])
             if new_volume == 0:
                 raise CommandSyntaxError("Please use PauseSong to mute the bot")
             elif not 0 < new_volume <= 100:
                 raise CommandSyntaxError("Expected value between 0 and 100")
-        except ValueError:
-            raise CommandSyntaxError(f"Value {new_volume} is not a valid integer")
         except IndexError:
             await respond(msg, f"**ANALYSIS: Current volume is {int(player.volume * 100)}%.**")
             return
+        except ValueError:
+            raise CommandSyntaxError(f"Value {new_volume} is not a valid integer")
+        if temp_flag:
+            player.prev_volume = player.volume
         player.volume = new_volume / 100
         await respond(msg, f"**AFFIRMATIVE. Set volume to {new_volume}%.**")
 
-    @Command("SongMode", "Shuffle", "Repeat",
+    @Command("SongMode",
              doc="Tells the bot in what order the bot should play its queue, or prints the current mode if no mode "
                  "is specified.",
              syntax="[n/normal|rs/repeat_song|rq/repeat_queue|s/shuffle|sr/shuffle_repeat]",
@@ -255,10 +275,100 @@ class MusicPlayer(BasePlugin):
             player.song_mode = SongMode.SHUFFLE
         elif arg in ("sr", "shuffle_repeat", "random_repeat"):
             player.song_mode = SongMode.SHUFFLE_REPEAT
+            player.already_played.clear()
         else:
             raise CommandSyntaxError(f"Argument {arg} is not a valid mode")
         await respond(msg, f"**AFFIRMATIVE. Song mode changed to {player.song_mode}.**")
 
+    @Command("Shuffle",
+             doc="Enables or disables shuffling of song order.",
+             syntax="[true/yes/on/false/no/off",
+             category="music_player")
+    async def _shuffle_toggle(self, msg):
+        player = self.get_guild_player(msg)
+        self.check_user_permission(msg.author, player)
+        shuffle_on = player.song_mode in (SongMode.SHUFFLE, SongMode.SHUFFLE_REPEAT)
+        try:
+            arg = is_positive(msg.clean_content.split(None, 1)[1])
+        except IndexError:
+            await respond(msg, f"**ANALYSIS: Shuffle is currently {'enabled' if shuffle_on else 'disabled'}.**")
+            return
+        if arg:
+            if shuffle_on:
+                await respond(msg, f"**ANALYSIS: Shuffle is already enabled.**")
+            else:
+                if player.song_mode is SongMode.NORMAL:
+                    player.song_mode = SongMode.SHUFFLE
+                    await respond(msg, f"**AFFIRMATIVE. Shuffle enabled.**")
+                elif player.song_mode is SongMode.REPEAT_QUEUE:
+                    player.song_mode = SongMode.SHUFFLE_REPEAT
+                    await respond(msg, f"**AFFIRMATIVE. Shuffle enabled with repeat.**")
+                elif player.song_mode is SongMode.REPEAT_SONG:
+                    await respond(msg, "**NEGATIVE. Shuffle cannot be enabled at the same time as song repeat.**")
+        else:
+            if shuffle_on:
+                if player.song_mode is SongMode.SHUFFLE:
+                    player.song_mode = SongMode.NORMAL
+                    await respond(msg, f"**AFFIRMATIVE. Shuffle disabled.**")
+                elif player.song_mode is SongMode.SHUFFLE_REPEAT:
+                    player.already_played.clear()
+                    player.song_mode = SongMode.REPEAT_QUEUE
+                    await respond(msg, f"**AFFIRMATIVE. Shuffle disabled. Queue repeat still enabled.**")
+            else:
+                await respond(msg, "**ANALYSIS: Shuffle is not enabled.**")
+
+    @Command("Repeat",
+             doc="Modifies repeating mode of song queue.",
+             syntax="[song/one/queue/all/false/no/off]",
+             category="music_player")
+    async def _repeat_mode(self, msg):
+        player = self.get_guild_player(msg)
+        self.check_user_permission(msg.author, player)
+        try:
+            arg = msg.clean_content.split(None, 1)[1].lower()
+            if arg not in ("song", "one", "queue", "all"):
+                arg = is_positive(arg)
+                if arg:
+                    raise CommandSyntaxError
+        except IndexError:
+            if player.song_mode is SongMode.REPEAT_SONG:
+                repeat_mode = "set to repeat current song"
+            elif player.song_mode is SongMode.REPEAT_QUEUE:
+                repeat_mode = "set to repeat the queue"
+            elif player.song_mode is SongMode.SHUFFLE_REPEAT:
+                repeat_mode = "set to repeat the queue. Shuffle is also enabled"
+            else:
+                repeat_mode = "disabled"
+            await respond(msg, f"**ANALYSIS: Repeat is currently {repeat_mode}.**")
+            return
+        if arg in ("song", "one"):
+            if player.song_mode is SongMode.REPEAT_SONG:
+                await respond(msg, "**ANALYSIS: Song repeat is already enabled.**")
+            elif player.song_mode in (SongMode.SHUFFLE, SongMode.SHUFFLE_REPEAT):
+                await respond(msg, "**NEGATIVE. Song repeat cannot be enabled at the same time as shuffle.**")
+            else:
+                player.song_mode = SongMode.REPEAT_SONG
+                await respond(msg, "**AFFIRMATIVE. Repeat mode set to song.**")
+        elif arg in ("queue", "all"):
+            if player.song_mode in (SongMode.REPEAT_QUEUE, SongMode.SHUFFLE_REPEAT):
+                await respond(msg, "**ANALYSIS: Queue repeat is already enabled.**")
+            elif player.song_mode is SongMode.SHUFFLE:
+                player.song_mode = SongMode.SHUFFLE_REPEAT
+                await respond(msg, "**AFFIRMATIVE. Repeat mode set to queue with shuffle.**")
+            else:
+                player.song_mode = SongMode.REPEAT_QUEUE
+                await respond(msg, "**AFFIRMATIVE. Repeat mode set to queue.**")
+        else:
+            if player.song_mode in (SongMode.NORMAL, SongMode.SHUFFLE):
+                await respond("**ANALYSIS: Repeat mode is disabled.**")
+            elif player.song_mode is SongMode.SHUFFLE_REPEAT:
+                player.already_played.clear()
+                player.song_mode = SongMode.SHUFFLE
+                await respond(msg, "**AFFIRMATIVE. Repeat mode disabled. Shuffle still enabled.**")
+            else:
+                player.song_mode = SongMode.NORMAL
+                await respond(msg, "**AFFIRMATIVE. Repeat mode disabled.**")
+
     @Command("SkipSong", "Skip",
              doc="Tells the bot to skip the currently playing song.",
              category="music_player")
@@ -299,6 +409,7 @@ class MusicPlayer(BasePlugin):
             await respond(msg, "**ANALYSIS: No music currently playing.")
             return
         player.queue.clear()
+        player.already_played.clear()
         player.stop()
         await respond(msg, "**ANALYSIS: The music has been stopped and the queue has been cleared.**")
 
@@ -416,9 +527,11 @@ class GuildPlayer:
         self.voice_client = voice_client
         self.logger = logging.getLogger(f"red_star.plugin.music_player.player_{self.voice_client.guild.id}")
         self.queue = []
+        self.already_played = set()
         self.is_playing = False
         self.current_song = {}
         self.song_mode = SongMode.NORMAL
+        self.prev_volume = None
         self._volume = self.parent.plugin_config["default_volume"] / 100
         self._loop = get_event_loop()
         self._song_start_time = None
@@ -427,57 +540,75 @@ class GuildPlayer:
         self.gid = str(voice_client.guild.id)
         self._alone_time = 0
 
-    async def enqueue(self, url):
+    async def enqueue(self, urls):
         with self.text_channel.typing():
+            # Fetch video info
             with YoutubeDL(self.parent.ydl_options) as ydl:
-                try:
-                    pl_slice = re.match(r"([^{`]+)`*(?:{([^}]*)})?", url)
-                    vid_info = await self._loop.run_in_executor(None, partial(ydl.extract_info, pl_slice[1],
-                                                                              download=False))
-                except YoutubeDLError as e:
-                    await self.text_channel.send(f"**WARNING. An error occurred while downloading video <{url}>. "
-                                                 f"It will not be queued.\nError details:** `{e}`")
-                    return
-            if vid_info.get("_type") == "playlist":
-                if pl_slice[2]:
-                    pl_slices = pl_slice[2].split(",")
-                    pl = []
-                    for s in pl_slices:
-                        if "-" in s:
-                            s_start, s_end = s.split("-", 1)
-                            try:
-                                s_start = int(s_start) - 1
-                            except ValueError:
-                                s_start = 0
-                            try:
-                                s_end = int(s_end)
-                                if s_end > len(vid_info["entries"]):
-                                    raise ValueError
-                            except ValueError:
-                                s_end = len(vid_info["entries"])
-                            if s_end < s_start:
-                                continue
-                            pl.extend(vid_info["entries"][s_start:s_end])
+                # Create a list of videos to be queued
+                to_queue = []
+                for url in urls:
+                    try:
+                        pl_slice = re.match(r"([^{`]+)`*(?:{([^}]*)})?", url)
+                        vid_info = await self._loop.run_in_executor(None, partial(ydl.extract_info, pl_slice[1],
+                                                                                  download=False))
+                    except YoutubeDLError as e:
+                        await self.text_channel.send(f"**WARNING. An error occurred while downloading video <{url}>. "
+                                                     f"It will not be queued.\nError details:** `{e}`")
+                        continue
+                    # If it's a playlist, we need to flatten it into our to_queue list, and deal with the slicing
+                    if vid_info.get("_type") == "playlist":
+                        # Slice handling
+                        if pl_slice[2]:
+                            pl_slices = pl_slice[2].split(",")
+                            for s in pl_slices:
+                                # Ranges handling
+                                if "-" in s:
+                                    s_start, s_end = s.split("-", 1)
+                                    try:
+                                        s_start = int(s_start) - 1
+                                    except ValueError:
+                                        s_start = 0
+                                    try:
+                                        s_end = int(s_end)
+                                        if s_end > len(vid_info["entries"]):
+                                            raise ValueError
+                                    except ValueError:
+                                        s_end = len(vid_info["entries"])
+                                    if s_end < s_start:
+                                        continue
+                                    to_queue.extend(vid_info["entries"][s_start:s_end])
+                                # Single values
+                                else:
+                                    try:
+                                        s = int(s) - 1
+                                        to_queue.append(vid_info["entries"][s])
+                                    except (ValueError, IndexError):
+                                        continue
+                        # No slice
                         else:
-                            try:
-                                s = int(s) - 1
-                                pl.append(vid_info["entries"][s])
-                            except (ValueError, IndexError):
-                                continue
-                else:
-                    pl = vid_info["entries"]
-                self._loop.create_task(self._enqueue_playlist(pl))
+                            to_queue.extend(vid_info["entries"])
+                    # If it's just a video, throw it in
+                    else:
+                        to_queue.append(vid_info)
+            # If we have more than one video to queue, toss it to the other function
+            if len(to_queue) > 1:
+                self._loop.create_task(self._enqueue_playlist(to_queue))
                 return
+            # Otherwise, special handling for common case of single videos
             else:
+                vid_info = to_queue[0]
+                # Check if queue is full
                 if len(self.queue) >= get_guild_config(self.parent, self.gid, "max_queue_length"):
                     await self.text_channel.send(f"**WARNING: The queue is full. {vid_info['title']} "
                                                  f"will not be added.")
                     return
+                # Check max duration limits
                 elif vid_info.get("duration", 0) > get_guild_config(self.parent, self.gid, "max_video_length"):
-                    max_len = pretty_duration(get_guild_config(self.parent, self.gid, "max_video_length"))
+                    max_len = fixed_width_duration(get_guild_config(self.parent, self.gid, "max_video_length"))
                     await self.text_channel.send(f"**WARNING: Your video exceeds the maximum video length "
                                                  f"({max_len}). It will not be added.**")
                     return
+                # Print time until song and start queueing
                 else:
                     time_until_song = ""
                     if len(self.queue) > 0:
@@ -496,37 +627,40 @@ class GuildPlayer:
         time_until_song = ""
         if len(self.queue) > 0:
             time_until_song = self.queue_duration + (self.current_song.get("duration", 0) - self.play_time)
-            time_until_song = f"\nTime until your song: {pretty_duration(time_until_song)}."
+            time_until_song = f"\nTime until your song: {pretty_duration(time_until_song)}"
         await self.text_channel.send(f"**ANALYSIS: Attempting to queue {len(entries)} videos. Your playback will "
                                      f"begin shortly.{time_until_song}**")
         orig_len = len(self.queue)
         with self.text_channel.typing():
-            for url in entries:
-                with YoutubeDL(self.parent.ydl_options) as ydl:
+            with YoutubeDL(self.parent.ydl_options) as ydl:
+                for url in entries:
                     try:
-                        vid_info = await self._loop.run_in_executor(None,
-                                                               partial(ydl.extract_info, url["url"], download=False))
+                        vid_info = await self._loop.run_in_executor(None, partial(ydl.extract_info, url["url"],
+                                                                                  download=False))
+                    # Skip broken videos while trying the rest
                     except YoutubeDLError as e:
                         await self.text_channel.send(f"**WARNING. An error occurred while downloading video "
                                                      f"<{url['url']}>. It will not be queued.\nError details:** `{e}`")
                         continue
-                if len(self.queue) >= get_guild_config(self.parent, self.gid, "max_queue_length"):
-                    await self.text_channel.send(f"**WARNING: The queue is full. No more videos will be added.**")
-                    break
-                elif vid_info.get("duration", 0) > get_guild_config(self.parent, self.gid, "max_video_length"):
-                    max_len = pretty_duration(get_guild_config(self.parent, self.gid, "max_video_length"))
-                    await self.text_channel.send(f"**WARNING: Video {vid_info['title']} exceeds the maximum video"
-                                                 f" length ({max_len}). It will not be added.**")
-                    continue
-                try:
-                    await self._process_video(vid_info)
-                except TypeError:
-                    continue
-                if not self.is_playing:
+                    # Abort once the queue is full
+                    if len(self.queue) >= get_guild_config(self.parent, self.gid, "max_queue_length"):
+                        await self.text_channel.send(f"**WARNING: The queue is full. No more videos will be added.**")
+                        break
+                    # Skip over videos that are too long
+                    elif vid_info.get("duration", 0) > get_guild_config(self.parent, self.gid, "max_video_length"):
+                        max_len = pretty_duration(get_guild_config(self.parent, self.gid, "max_video_length"))
+                        await self.text_channel.send(f"**WARNING: Video {vid_info['title']} exceeds the maximum video"
+                                                     f" length ({max_len}). It will not be added.**")
+                        continue
                     try:
-                        self._loop.create_task(self._play())
-                    except ClientException:
-                        pass
+                        await self._process_video(vid_info)
+                    except TypeError:
+                        continue
+                    if not self.is_playing:
+                        try:
+                            self._loop.create_task(self._play())
+                        except ClientException:
+                            pass
         await self.text_channel.send(f"**ANALYSIS: Queued {len(self.queue) - orig_len} videos.**")
         if get_guild_config(self.parent, self.gid, "print_queue_on_edit") and self.queue:
             final_msg = f"**ANALYSIS: Current queue:**{self.print_queue()}"
@@ -558,7 +692,15 @@ class GuildPlayer:
         try:
             if self.song_mode == SongMode.SHUFFLE_REPEAT:
                 try:
-                    next_song = self.queue[randint(0, len(self.queue) - 1)]
+                    vids = {i for i, _ in enumerate(self.queue)}
+                    eligible_songs = vids - self.already_played
+                    if not eligible_songs:
+                        self.already_played.clear()
+                        eligible_songs = vids
+                    eligible_songs = tuple(eligible_songs)
+                    next_song = eligible_songs[randint(0, len(eligible_songs) - 1)]
+                    self.already_played.add(next_song)
+                    next_song = self.queue[next_song]
                 except ValueError:
                     next_song = self.queue[0]
             elif self.song_mode == SongMode.SHUFFLE:
@@ -595,6 +737,10 @@ class GuildPlayer:
             file = next_song["url"]
             before_args += " -reconnect 1 -reconnect_streamed 1 -reconnect_delay_max 30"
         self.is_playing = True
+        if self.prev_volume:
+            self.volume = self.prev_volume
+            self.prev_volume = None
+            await self.text_channel.send(f"**ANALYSIS: Volume reset to {int(self.volume * 100)}%.**")
         source = PCMVolumeTransformer(FFmpegPCMAudio(file, before_options=before_args, options="-vn"),
                                       volume=self._volume)
         self.voice_client.play(source, after=self._after)
@@ -683,7 +829,7 @@ class GuildPlayer:
     def print_queue(self):
         str_list = []
         for i, vid in enumerate(self.queue):
-            duration = pretty_duration(vid.get("duration", 0))
+            duration = fixed_width_duration(vid.get("duration", 0))
             # Title truncating and padding
             title = vid.get("title", "Unknown")
             if len(title) >= 59:
@@ -693,7 +839,7 @@ class GuildPlayer:
 
     def print_now_playing(self):
         play_time, duration, progress = self.progress
-        dur_str = pretty_duration(play_time) + "/" + pretty_duration(duration)
+        dur_str = fixed_width_duration(play_time) + "/" + fixed_width_duration(duration)
         progbar = progress_bar(progress, 70)
         title = self.current_song.get("title", "Unknown")
         if len(title) > 57:
@@ -722,11 +868,23 @@ def seconds_to_minutes(secs, hours=False):
 
 
 def pretty_duration(seconds):
+    if seconds <= 0:
+        return "Unknown"
+    elif seconds >= 3600:
+        hours, minutes, seconds = seconds_to_minutes(seconds, hours=True)
+        return f"{hours}:{minutes:02d}:{seconds:02d}"
+    else:
+        minutes, seconds = seconds_to_minutes(seconds)
+        return f"{minutes:02d}:{seconds:02d}"
+
+
+def fixed_width_duration(seconds):
     minutes, seconds = seconds_to_minutes(seconds)
     if seconds <= 0:
         return "--:--"
     if minutes > 99:
-        minutes += 1
+        if seconds >= 30:
+            minutes += 1
         return f"{minutes:04d}m"
     else:
         return f"{minutes:02d}:{seconds:02d}"
